/*
								 ***********
								** Summary **
								 ***********

					  Copyright 2001, Nagual Enterprises
				Please see COPYING for the complete GNU license

	(1)	Title:
				TES:2 Daggerfall SaveTree Classes

	(2)	Purpose:
			To supply the data structures needed to access the
			contents of a Daggerfall saved game.

	(3)	Change history:
			Please see the RCS logs.

	(4)	Definitions:

			Constants:
				chunk_				--	Include already processed flag.
				ChunkXXX			--	Manifest constants for the variable
										block types in savetree.dat.
				DirSep				--	Pathname component separator.
				QbnFactionOffset	--	Offset within a chunk where faction
										types are coded for some chunk types.
				SavePrefix			--	Saved game directory prefix.
				SaveFile			--	Saved game file name.
				SlotMax				--	Number of saved game slots.

			Macros:
				punt				--	Write a period-newline pair on a stream.

			Classes:
				Arena2DataHeader	--	The common node structure in every
										branch of the saved tree image.
				Chunk				--	Mapping for a segment's data elements.
				ChunkHeader			--	Mapping for a segment descriptor.
				ColumnStream		--	ostrstream/ostream subclass to implement
										news paper column-type output.
				SaveTree			--	File stream subclass to manage access to
										a saved game.

			Assertions about definitions:
				Note that some implementations of C++ impose alignment
				restrictions on C++ classes and so we prefer to use
				structs in those cases where that imposed alignment may
				clash with the raw data streams being modelled.  That way,
				sizeof() and [] do the right thing without juggling the
				books.

	(5)	References:

			Include files:
				file.h				--	Various ios shorthands.

			Methods:
				Please see the class scaffold for methods.

			Classes:
				***  None  ***

			Assertions about the references:
				***  None  ***

								 ********
								** Text **
								 ********
*/

#ifndef chunk_
#	define chunk_

#	define SlotMax			5
#	define SavePrefix		"save"
#	define SaveFile			"savetree.dat"

#	ifdef __MSDOS__
#		define DirSep		"\\"
#	else
#		define DirSep		"/"
#	endif
#	include "file.h"

		/*	Recent libc's are broken and return the file size as the
			block size on linux. */

#	ifdef linux
#		define fileSize(statSt)		(statSt.st_blksize)
#	else
#		define fileSize(statSt)		(statSt.st_size)
#	endif

	inline ostream& punt
	(	ostream& os)
	{	return os << '.' << endl;}

	/*	SaveTree.dat format:

		fixed header char[0x13];
		Thereafter, counted chunks starting at file offset 13:

			long chunkSize;
			byte chunk[chunkSize];

		The 4 bytes for storing the chunk size are not counted by the
		chunk size itself.

		chunk[0] is the chunk type code.

		0x1f is a magic potion record.
		0x0a is an active spell record preamble.
		0x09 is a magic spell record.
		0x03 is the player record chunk.
		0x02 is an inventory object record.

		Note that the chunk size can be 0, in which case the next chunk size
		begins at the present (or next) offset of the file.

		Each saved game may have 3 zero-sized chunks delimiting
		sections of the file:
			o  optional untyped town chunk
			   (always the first chunk if present)
			o  map grid points and player possessions
			   (most chunks are here)
			o  Quest-related chunks
			   (essentially always present)
			o  optional untyped dungeon chunk
			   (always the last chunk if present) */

	/*	Established chunk types.  Types tagged as `unknown' have been
		encountered, but aren't presently well-known. */

#	define ChunkItem			'\x02'
#	define ChunkPlayer			'\x03'
#	define ChunkPlayerLocation	'\x04'
#	define ChunkPlayerParent	'\x05'
#	define ChunkUnknown_06		'\x06'
#	define ChunkUnknown_07		'\x07'
#	define ChunkUnknown_08		'\x08'
#	define ChunkSpell			'\x09'
#	define ChunkGuild			'\x0a'
#	define ChunkOriginalStats	'\x0b'
#	define ChunkQuest			'\x0e'
#	define ChunkQuestParent		'\x10'
#	define ChunkQbnFoe			'\x12'
#	define ChunkMagicFoe		'\x16'
#	define ChunkControlPanel	'\x17'
#	define ChunkLogbook			'\x18'
#	define ChunkBankAccount		'\x19'
#	define ChunkOriginalClass	'\x1c'
#	define ChunkUnknown_1d		'\x1d'
#	define ChunkPotion			'\x1f'
#	define ChunkLocation		'\x20'
#	define ChunkTempTreasure	'\x21'
#	define ChunkCreature1		'\x22'
#	define ChunkUnknown_24		'\x24'
#	define ChunkUnknown_27		'\x27'
#	define ChunkQbnPlace		'\x28'
#	define ChunkQbnQuestor		'\x29'
#	define ChunkUnknown_2a		'\x2a'
#	define ChunkUnknown_2b		'\x2b'
#	define ChunkCreature2		'\x2c'
#	define ChunkNPC				'\x2d'
#	define ChunkModelNPC		'\x2e'
#	define ChunkCurrentAMF		'\x33'
#	define ChunkContainer		'\x34'
#	define ChunkUnknown_35		'\x35'
#	define ChunkTreasure		'\x3a'
#	define ChunkUnknown_40		'\x40'
#	define ChunkUnknown_41		'\x41'

	typedef struct _ChunkHeader ChunkHeader;
	struct _ChunkHeader
	{
		long chunkSize;				//	Byte count of subsequent data record.
		unsigned char type;			//	Type of data chunk.
	};

	typedef struct _Chunk Chunk;
	struct _Chunk
	{
		long chunkSize;				//	Bytes in this data chunk.
		void* chunk;				//	Pointer to the data record.

		unsigned char type()		//	Answer the receiver's chunk type.
			{return chunk ?
				*(unsigned char*)chunk :
				'\0';}
	};

		/*	These data elements are in common with every record (well, at
			least the *interesting* ones) in the save file. */

	typedef struct _Arena2DataHeader Arena2DataHeader;
	struct _Arena2DataHeader
	{
		char key;						//	Data type code.
		char filler0a[6];
		long position[3];				//	E3 (x, y, z) for this item.
		char filler0b[8];
		unsigned short picture1;		//	Texture file:index (bits 9:7).
		unsigned short picture2;
		unsigned long objectId;			//	Hash code for this object.
										//	Hi half is often the permanent site
										//	key (0xc36a, etc)
		char filler1a[3];
		char questNo;					//	Quest #, if associated with quest.
		unsigned long parentId;			//	Hash code for this object's owner.
		char filler1[8];
		long questObject;				//	Quest object hash marker.
		long nextObject;
		long prevObject;
		long sublistHead;
		long index4;
	};

#	define QbnFactionOffset		(18 + sizeof(Arena2DataHeader))

	typedef struct _GenericChunk GenericChunk;
	struct _GenericChunk
	{
//		char key;
		Arena2DataHeader head;
	};


		/*	Prototype for operations to perform on the saved game data. */

	class SaveGameBrowser;
	typedef int (SaveGameBrowser::*ChunkIterator)(Chunk*);
	typedef int (SaveGameBrowser::**ChunkIteratorList)(Chunk*);


	class SaveTree
	{
/*
					 ********************
					** Instance methods **
					**  and  variables  **
					 ********************
*/

	protected:
		SaveGameBrowser* _browser;		//	Browser operating on current game.
		fstream* _handle;				//	Current saved game.
		const char* _name;				//	Saved game file name.
		int _isOpen;					//	True once the file is open.
		ChunkHeader _header;			//	Current header record.
		Chunk _currentChunk;			//	Current chunk record.
		long _offset;					//	Current record offset in file.

		Chunk* chunk(Chunk* chunk);		//	Answer the chunk at the current
										//	file position.
//		Chunk*							//	Find the next chunk header of the
//			findChunk(unsigned char t);	//	specified type.
		Chunk* nextChunk();				//	Answer the next chunk record.
		ChunkHeader*					//	Answer the chunk header following
			nextHeader(void);			//	the supplied header.
		int openChunk();				//	Open the chunk reader.

	public:

		SaveTree() :					//	Construct a save tree reader
			_browser(NULL),
			_handle(NULL),
			_name(NULL),
			_isOpen(false),
			_offset(0) {};
		SaveTree(fstream* f,			//	Construct a save tree reader
				 const char* n) :
			_browser(NULL),
			_handle(f),
			_name(n),
			_isOpen(false),
			_offset(0) {};
		~SaveTree() {};

		SaveGameBrowser* browser()		//	Answer the current browser for
			{return _browser;}			//	the save game.
		void browser(SaveGameBrowser* b)
			{_browser = b;}
		void close()
			{_handle->close();}
										//	Perform each action in the list
										//	for each chunk record in the
										//	receiver;.
		int forEach(ChunkIterator* item);
		fstream* handle()
			{return _handle;}
		const char* name(const char* n)	//	Assign the receiver's file name.
			{return _name = n;}
		const char* name()
			{return _name;}
		long offset()					//	Answer the current chunk offset.
			{return _offset;}
		long offset(long o)				//	Assign the current chunk offset.
			{return _offset = o;}
		int update(	long offset,		//	Update the specified chunk in place.
					Chunk* c);

/*
					 *****************
					** Class methods **
					** and variables **
					 *****************
*/

		static int OpenMode;
		static int NullCounts;
	};


		/*	Derive a new output string stream able to format strings into
			multiple columns.

			ColumnStreams are fixed-width columns with a fixed number
			of columns per line, with an associated output stream that
			receives the assembled line of text once the last column is
			filled.

			ColumnStreams have no notion of page length, so filling the
			line of a single column advances to the next column, not to
			the next line of the same column. */

		/*	Unfortunately ios inheritance is broken in recent GNU (2.91)
			releases. */

	class ColumnStream //: public ostrstream
	{
/*
					 ********************
					** Instance methods **
					**  and  variables  **
					 ********************
*/

	protected:
		ostream& associatedStream;		//	The associated output stream that
										//	records assembled lines.
		int columnWidth;				//	The size (in characters) of each
										//	column.
		int columnCount;				//	The number of columns per line.
		int currentColumn;				//	The current column used for output.
		char* associatedBuf;			//	Pointer to the string used to prepare
										//	each line of output.
		
	public:
		ostrstream basis;
		ColumnStream(char* s, int w) :	//	Stock strstream constructor.
			associatedStream(cout),
			columnWidth(40),
			columnCount(2),
			currentColumn(1),
			associatedBuf(s),
			basis(s, w) {};

		ostream& advance(void);			//	Advance the output stream to the next
										//	column.  Perform the requisite line
										//	break after the final column on a line.
		int column(int c)				//	Change the current column.
			{return currentColumn = c;}
		int column(void)				//	Answer the current output column.
			{return currentColumn;}
		int columns(int c)				//	Change the operating number of columns.
			{return columnCount = c;}
		int columns(void)				//	Answer the current columns / line.
			{return columnCount;}
		ostream& flushColumn(void);		//	Flush the assembled line to the
										//	associated stream.
		ostream& rewind(void);			//	Prepare the line buffer for reuse.
//		ostream&						//	Assign the ostream to record the
//			stream(ostream& os)			//	assembled columns on.
//			{return associatedStream = os;}
		int width(int w)				//	Change the column width for subsequent
			{return columnWidth = w;}	//	output lines.
		int width(void)					//	Answer the current column width.
			{return columnWidth;}
	};

/*
					 *******************
					** In-line methods **
					 *******************
*/

	inline ostream& ColumnStream::rewind
	(	void)
	{
		currentColumn = 1;
		basis.seekp(0, ios::beg);
		return basis; //(ostream&)*this;
	}

#endif					/*	chunk_ */
